import re

def parse_hgvs(hgvs):
    """
    Parse an hgvs (or hgvs-like) description of a tandem repeat unit

    :param hgvs: a string containing the variant description
    :return: a dictionary containing of the form
        {
            "start": start_position,
            "end": end_position,
            "unit": the repeat unit string
            "count": the number of repeats
        }
    """
    regex_string = "(?P<start>[0-9]+)_(?P<end>[0-9]+)(?P<unit>[A-Za-z]*)\[(?P<count>[0-9]+)\]"
    regex = re.compile(regex_string)

    match = regex.match(hgvs)

    assert match is not None, "description could not be parsed"
    if len(match.group("unit")) > 0:
        assert len(match.group("unit")) == int(match.group("end")) + 1 - int(match.group("start")), "range and sequence do not match"

    return {
        "start": int(match.group("start")),
        "end": int(match.group("end")),
        "unit": match.group("unit"),
        "count": int(match.group("count"))
    }


def str_search(repeat_unit, sequence, start_pos=0, offset=0):
    """
    Search a sequence for short tandem repeats of a canonical repeat unit.

    :param repeat_unit: The canonical repeat unit
    :param sequence:    The sequence to search within
    :param start_pos:   Position in 'sequence' at which to begin the search
    :param offset:      Optional integer offset to be added to the start position
    """
    regex_string = "({}){{2,}}".format(repeat_unit)
    regex = re.compile(regex_string)

    for result in regex.finditer(sequence, pos=start_pos):
        yield dict(
            position=result.span()[0] + offset,
            canonical_unit=repeat_unit,
            num_units=(result.span()[1] - result.span()[0]) // len(repeat_unit),
            repeat_sequence=result.group(0)
        )


def str_search_aligned(repeat_unit, sequence):
    """
    Search a (gapped) aligned sequence for short tandem repeats of a canonical repeat_unit

    :param repeat_unit: The canonical repeat unit
    :param sequence:    The (aligned) sequence to search within
    :return: a summary dict for each found repeat
    """
    regex_string = "(" + "".join(["{}[-]*".format(n) for n in repeat_unit]) + "){2,}"
    regex = re.compile(regex_string)
    for result in regex.finditer(sequence):
        gapped_seq = result.group(0)
        ungapped_seq = gapped_seq.replace("-", "")
        yield dict(
            start = result.span()[0],
            end = result.span()[1] - 1,
            canonical_unit = repeat_unit,
            num_units = len(ungapped_seq) // len(repeat_unit),
            gapped_sequence = gapped_seq,
            ungapped_sequence = ungapped_seq
        )


"""
Alignment of sequences
"""

from Bio import SeqIO, pairwise2
from Bio.Seq import reverse_complement
import edlib
from interval import interval
import numpy as np
import sys

# increase the max number of alignments that pairwise2 will return (default is 1000, which is insufficient when alignment is complex)
pairwise2.MAX_ALIGNMENTS = 10000


def get_edit_dist(seq1, seq2):
    """
    Get the edit distance between two sequences
    The edit distance is computed using edlib as it is considerably faster than python levenshtein
    
    :param seq1, seq2: sequences to be compared
    :type seq1, seq2:  string
    :return:           the edit distance (int)
    """
    return edlib.align(seq1, seq2, task="distance", mode="NW")["editDistance"]


def get_strand_dists(query_seq, ref_seq):
    """
    Measure the distances between a query sequence and reference sequence,
    and the reverse complement of query sequence and the reference sequence
    
    :param query_seq, ref_seq: query and reference sequences
    :type query_seq, ref_seq:  string
    :return:                   a tuple containing the pair of distances (d_fwd, d_rev) 
    """
    fwd = query_seq
    rev = reverse_complement(fwd)
    d_fwd = get_edit_dist(fwd, ref_seq)
    d_rev = get_edit_dist(rev, ref_seq)
    return (d_fwd, d_rev)
    
    
def adjust_strand_dir(query_seq, ref_seq):
    """
    Adjust the strandedness of 'query_seq' to match that of 'ref_seq'
    This is achieved by determining the edit distance between query_seq and its reverse_complement
    to 'ref_seq', and choosing the direction which gives the smallest distance
        
    :param ref_seq:   The reference sequence
    :type ref_seq:    string
    :param query_seq: The query sequence
    :type query_seq:  string
    :param verbose:   A flag indicating if results should be printed to console
    :return:          A SeqRecord containing the query sequence adjusted to match 
                      the reference direction
    """
    d_fwd, d_rev = get_strand_dists(query_seq, ref_seq)
   
    if d_rev < d_fwd:
        return reverse_complement(query_seq)

    return query_seq


def find_indels(ref_aln, qry_aln):
    """
    Identify the positions of indels in the alignment of 2 sequences
    
    :param ref_aln: An aligned sequence representing the reference
    :param qry_aln: An aligned sequence representing the query
    :yields:        The position of each gap character in the alignment
    """
    assert len(ref_aln) == len(qry_aln), "Error - Aligned sequence lengths do not match"
    for pos, pair in enumerate(zip(ref_aln, qry_aln)):
        if '-' in pair:
            yield pos
            

def pick_leftmost(alignments):
    """
    Given a list of alignments, identify the most left-justified (i.e. gaps are rightmost)
    
    To identify the rightmost positioning of gaps, the mean of the gap end positions
    for each alignment is used as a heuristic.
    
    :param alignments: A list of alignment tuples with structure (ref, query, score)
    :return:           The highest scoring left-justified alignment
    """
    best_score = 0
    best_alignment = alignments[0]
    for aln in alignments:
        gaps = []
        gap = None
        for i in range(len(aln[0])):
            if aln[0][i] == '-' or aln[1][i] == '-':
                if gap is None:
                    gap = []
                    gap.append(i) # gap start
            else:
                if gap is not None:
                    gap.append(i-1) # gap end
                    gaps.append(gap)
                    gap = None
        
        if gap is not None:
            gap.append(i-1) # gap end
            gaps.append(gap)
        
        if len(gaps) == 0:
            break
            
        score = np.mean([g[-1] for g in gaps])
        if score > best_score:
            best_score = score
            best_alignment = aln
           
    return best_alignment


def parse_cigar(cigar):
    """
    Parse a cigar string into a list of operations
    
    :param cigar: a cigar string
    :returns:     A list of tuples of the form (length, operation)
    """
    cigar = cigar
    ops = []
    while len(cigar) > 0:
        for i, c in enumerate(cigar):
            if c in ("=", "X", "D", "I"):
                op = cigar[:i+1]
                ops.append((int(op[:-1]), op[-1]))
                cigar = cigar[i+1:]
                break
    return ops


def apply_cigar(query, target, cigar):
    """
    Given two unaligned sequences and a cigar string, return 
    aligned sequences
    
    :param query:  query sequence
    :param target: regerence sequence
    :param cigar:  cigar string
    :returns:      a dictionary containing query and target sequences
                   with gap characters
    """
    ops = parse_cigar(cigar)
    ori1, ori2 = (query, target)
    seq1, seq2 = ("", "")
    
    for count, op in ops:
        if op in ("=", "X"):
            seq1 += ori1[:count]
            ori1 = ori1[count:]
            seq2 += ori2[:count]
            ori2 = ori2[count:]
        elif op == "I":
            seq1 += ori1[:count]
            ori1 = ori1[count:]
            seq2 += "-" * count
        elif op == "D":
            seq1 += "-" * count
            seq2 += ori2[:count]
            ori2 = ori2[count:]
    
    return {"query": seq1, "target": seq2}


default_alignment_opts = dict(
    match=3,
    mismatch=-3,
    gap_open=-9,
    gap_extend=-2,
    penalize_opening=True,
    semi_global=True
)

def get_alignment(ref_seq, query_seq, aln_opts={}):
    """
    Perform a (semi) global alignment of two sequences using bio.pairwise2
    
    :param ref_seq:   reference sequence (str)
    :param query_seq: query sequence (str)
    :param aln_opts:  additional arguments to be passed to pairwise2.align.globalms
        
    :return: The top scoring, most left-aligned alignment
    """
    opts = default_alignment_opts.copy()
    opts.update(aln_opts)
    
    alns = pairwise2.align.globalms(
        ref_seq, query_seq,
        opts["match"],
        opts["mismatch"],
        opts["gap_open"],
        opts["gap_extend"],
        penalize_end_gaps=not opts["semi_global"],
        penalize_extend_when_opening=opts["penalize_opening"]
    )
    
    # print out a warning if max_alignments is reached, as in this case correct justification is not guaranteed
    if len(alns) == pairwise2.MAX_ALIGNMENTS:
        print("Warning: MAX_ALIGNMENTS ({}) exceeded".format(pairwise2.MAX_ALIGNMENTS), file=sys.stderr)
    
    return pick_leftmost(alns)


def align(ref_seq, query_seq, window=100, aln_opts={}):
    """
    Align two sequences then perform realignment around indels
    in order to obtain consistent gap positions.
    
    1. Coarse alignment
    2. Identify the location of indels
    3. Create intervals of 'window' size around the indel locations
    4. Merge overlapping intervals
    5. Fine realignment of the sequences contained within each interval
    6. Splice the re-aligned sections back into the original alignment
    
    :param ref_seq:   The reference sequence
    :param query_seq: The query sequence
    :param window:    The number of nucleotides each side of an indel to use for re-alignment
    :param aln_opts:  Options to be passed to the alignment function
                      See variant_tools.alignment.get_alignment
    
    :returns:         A tuple containing the aligned reference sequence, the aligned query sequence
                      and the intervals used for re-alignment
    """
    
    # adjust strandedness
    query_seq = adjust_strand_dir(query_seq, ref_seq)
    
    # coarse align
    aln = edlib.align(query_seq, ref_seq, mode="NW", task="path")
    result = apply_cigar(query_seq, ref_seq, aln["cigar"])
    
    # use aligned sequences
    query_seq = result["query"]
    ref_seq = result["target"]
    
    aln_len = len(ref_seq)
    intervals = interval()
    
    # create intervals which cover the indel sites
    for pos in find_indels(ref_seq, query_seq):
        intervals |= interval[max(0, pos - window), min(aln_len, pos + window)]

    # strings to hold the patched alignments
    new_ref_seq = ""
    new_query_seq = ""
    
    # the last position covered
    last_pos = 0
    
    # re-align each interval
    for i in intervals.components:
        # start and end of the interval
        start, end = int(i[0][0]), int(i[0][1])
        
        # add the original sequences from last used position up to current start
        new_ref_seq += ref_seq[last_pos:start]
        new_query_seq += query_seq[last_pos:start]
        
        # Extract the aligned sequences within the current interval 
        # and remove any gap characters
        r = ref_seq[start:end].replace('-', '')
        q = query_seq[start:end].replace('-', '')
        
        # re-align
        aln = get_alignment(r, q, aln_opts)
        
        # splice the realigned section back into the original alignment
        new_ref_seq += aln[0]
        new_query_seq += aln[1]
        
        last_pos = end
    
    # add any remaining sequence
    new_ref_seq += ref_seq[last_pos:]
    new_query_seq += query_seq[last_pos:]
   
    return (new_ref_seq, new_query_seq, intervals)
    

def pretty_alignment(seq1, seq2, start=0, end=None, width=200, match=" ", mismatch="*", gap="#"):
    """
    Pretty string representation of the alignment
    :param seq1, seq2: aligned sequence strings
    :param start:      the alignment position at which to start
    :param end:        the alignment position at which to end
    :param width:      the number of characters per line
    :param match:      character used to represent a match
    :param mismatch:   character used to represent a mismatch
    :param gap:        character used to represent a gap
    """
    pos = start
    if end is None:
        end = len(seq1)
    
    def match_symbol(base1, base2):
        if base1 == base2:
            return match
        if base1 == "-" or base2 == "-":
            return gap
        return mismatch
        
    pretty = []

    while pos < end:
        num_bases = width if pos + width < end else end-pos
        pretty.append(seq1[pos: pos + num_bases])
        pretty.append("".join([match_symbol(seq1[pos+i], seq2[pos+i]) for i in range(num_bases)]))
        pretty.append(seq2[pos:pos+num_bases])
        pretty.append("")
        pos += width
        
    return "\n".join(pretty)



"""
Utilities for vcf records
"""

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from .mutalyzer import chrom_to_accession, apply_variants
from .utils import HGVS


def generate_accession_dict(reader, build='hg19'):
    """
    Generate dictionary of chrom -> accession
    for all chromosomes in reader
    :param reader: cyvcf2 reader
    :param build: genome build
    :return: dict of chrom -> accession
    """
    return {c: chrom_to_accession(c, build) for c in reader.seqnames}


def vcf_to_hgvs(record, sample_idx, accession_dict=None):
    """
    Create delins hgvs representation of a vcf record. 
    Will select the allele(s) present in the given sample
    
    :param record: cyvcf2 record
    :param sample_idx: sample index
    :param accession_dict: optional dictionary of chrom -> accession nr

    :return: A list of hgvs objects.
    List may be empty if sample is homozygous reference for the site.
    """
    gts = record.gt_bases[sample_idx]
    if "|" in gts:
        bases = gts.split("|")
    else:
        bases = gts.split("/")
    true_alt = set(bases) - set([str(record.REF)])
    if len(str(record.REF)) == 1:
        pos_str = "{0}".format(record.POS)
    else:
        pos_str = "{0}_{1}".format(record.POS, record.POS + len(str(record.REF)) - 1)
    hgvs = []
    for alt in true_alt:
        if alt == "." or alt == "*" or alt is None:
            continue
        if accession_dict is not None:
            chrom = accession_dict[record.CHROM]
        else:
            chrom = record.CHROM
        desc_str = "{0}del{1}ins{2}".format(pos_str, record.REF, alt)
        hg = HGVS(chrom, "g", desc_str)
        hgvs.append(hg)
    return hgvs


def region_to_hgvs(reader, region, sample_idx, accession_dict=None):
    """
    Generator of hgvs descriptions for a region
    :param reader: cyvcf2 reader
    :param region: Region object
    :param sample_idx: sample index
    :param accession_dict: optional dictionary of chrom -> accession nr
    :return: generator of hgvs descriptions
    """
    region_str = "{0}:{1}-{2}".format(
        region.chromosome,
        region.start,
        region.end
    )
    for record in reader(region_str):
        for hg in vcf_to_hgvs(record, sample_idx, accession_dict):
            yield hg


def region_to_sequence(reader, region, sample_idx, accession_dict=None):
    """
    Create applied sequence from vcf region
    :param reader: cyvcf2 reader
    :param region: Region object
    :param sample_idx: sample index
    :param accession_dict: optional dictionary of chrom -> accession nr
    :return: string containing sequence
    """
    hgs = [x.description for x in region_to_hgvs(reader, region, sample_idx, accession_dict)]
    if accession_dict is not None:
        if region.chromosome not in accession_dict:
            accession_dict.update({region.chromosome: chrom_to_accession(region.chromosome)})
        return apply_variants(hgs, accession_dict[region.chromosome], region.start, region.end)
    return apply_variants(hgs, region.chromosome, region.start, region.end)


def sequence_to_fasta(sequence, name):
    """
    Create fasta string of a sequence and name
    :param sequence: string with sequence
    :param name: name of sequence
    :return: fasta string. This _will_ contain newlines
    """
    s = Seq(sequence)
    r = SeqRecord(s, id=name, description="")
    return r.format("fasta")


def region_to_fasta(reader, region, sample_idx, accession_dict=None):
    """
    Create fasta string from vcf region

    The name of the fasta sequence will be the region's value attribute.
    If this attribute is None, the string representation of the region
    will be the fasta record name.
    :param reader: cyvcf2.reader
    :param region: Region object
    :param sample_idx: sample index
    :param accession_dict: optional dictionary of chrom -> accession nr
    :return: fasta string. Contains newlines.
    """
    seq = region_to_sequence(reader, region, sample_idx, accession_dict)
    if region.value is not None:
        return sequence_to_fasta(seq, region.value)
    else:
        return sequence_to_fasta(seq, str(region))



"""
Generic utils
"""

from collections import namedtuple


class Region(namedtuple('Region', ['chromosome', 'start', 'end',  'value'])):
    __slots__ = ()

    def __new__(cls, chromosome, start, end, value=None):
        return super().__new__(cls, chromosome, start, end, value)

    def __str__(self):
        return "{0}:{1}-{2}".format(self.chromosome, self.start, self.end)


class HGVS(namedtuple("HGVS", ['accesssion', 'type', 'description'])):

    def __str__(self):
        return "{0}:{1}.{2}".format(self.accesssion, self.type, self.description)


def regions_from_file(inputf):
    """Generator of Region's from BED file"""
    with open(inputf) as handle:
        for line in handle:
            contents = line.strip().split("\t")
            chrom, start, end = contents[:3]
            if len(contents) == 3:
                value = None
            else:
                value = contents[3]
            yield Region(chrom, int(start), int(end), value)



import re

import click
import cyvcf2

from .utils import Region, regions_from_file
from .vcf_tools import generate_accession_dict, region_to_fasta


def validate_region_str(ctx, param, value):
    """
    Validate region string as a click parameter. 
    Must be of format 'chromosome:start-end'
    Returns region object
    May raise BadParameter.
    """
    if value is None:
        return None
    region_re = re.compile('^([\w\d]+):(\d+)-(\d+)$')
    match = region_re.match(value)
    if match is not None:
        return Region(match.group(1),
                      int(match.group(2)),
                      int(match.group(3)))
    else:
        raise click.BadParameter('{0} is not a '
                                 'valid region string'.format(value))


@click.command(short_help="Convert vcf records to fasta sequence")
@click.option("--vcf", "-v", type=click.Path(exists=True), required=True,
              help="Path to input VCF")
@click.option("--sample", "-s", type=click.STRING,
              help="Sample to consider. "
                   "If not given, take fist sample in VCF")
@click.option("--region", "-R", callback=validate_region_str,
              help="Region string of format chr:start-end")
@click.option('--region-file', "-L", type=click.Path(exists=True),
              help="Path to BED file containing regions")
def vcf2sequence(vcf, sample=None, region=None, region_file=None):
    """
    Convert vcf records into fasta sequences.

    Will emit a fasta record for every region in `-L` or `-R`.
    Names of sequences will be taken from either:
        - an optional fourth column in the BED file
        - the string representation of the region
    """
    if region is None and region_file is None:
        raise ValueError("Either -R or -L must be set")

    reader = cyvcf2.VCF(vcf)
    if sample is not None:
        sample_idx = reader.samples.index(sample)
    else:
        sample_idx = 0

    acdict = generate_accession_dict(reader)

    if region:
        print(region_to_fasta(reader, region, sample_idx, acdict), end='')

    if region_file:
        for reg in regions_from_file(region_file):
            print(region_to_fasta(reader, reg, sample_idx, acdict), end='')



"""
Utilities using the Mutalyzer web service
"""
import requests
from collections import defaultdict
import re

# cache for mutalyzer sliceChromosome results
chr_id_cache = defaultdict(str)

# url to the mutalyzer api
MUTALYZER = "https://mutalyzer.nl"


def offset_hgvs(hgvs, offset):
    """
    change the positions associated with a (simple) variant in hgvs notation
    by applying the specified offset

    :param hgvs: A string containing a variant description in hgvs notation
    :param offset: An integer offset to be applied to variant coordinates
    :return: A string containing the hgvs variant description after offsetting
    """
    if '_' in hgvs:
        # coordinate pair
        m = re.match(r'([0-9]+)[_]+([0-9]+)([a-zA-Z]+.*)', hgvs, re.M | re.I)
        assert len(m.groups()) == 3, "could not decode variant"
        return str(int(m.group(1)) + offset) + "_" + str(int(m.group(2)) + offset) + m.group(3)
    else:
        # single coordinate
        m = re.match(r'([0-9]+)([a-zA-Z]+.*)', hgvs, re.M | re.I)
        assert len(m.groups()) == 2, "could not decode variant"
        return str(int(m.group(1)) + offset) + m.group(2)


def get_mutalyzer_region_id(chrAcc, start, end):
    """
    Retrieve an ID from mutalyzer sliceChromosome SOAP service which represents
    the target region of the reference genome.
    The ID can be used as a reference against which to apply other 
    mutalyzer operations

    :param chrAcc: chromosome accession number
    :param start: start position of region (1-based inclusive)
    :param end: end position of region (1-based inclusive)
    :return: a string containing the mutalyzer id
    """

    # if this region is cached, use the cached result
    cached_id = chr_id_cache[(chrAcc, start, end)]
    if cached_id != "":
        return cached_id

    # otherwise query mutalyzer
    mutalyzer_url = "{}/json/sliceChromosome".format(MUTALYZER)
    mutalyzer_params = {
        "chromAccNo": chrAcc,
        "start": str(start),
        "end": str(end),
        "orientation": "1" 
    }

    #print("Querying mutalyzer for chromosome region")
    response = requests.get(url=mutalyzer_url, params=mutalyzer_params)
    ref_id = response.text.strip('"')
    #print("Mutalyzer returned {}".format(ref_id))
    assert ref_id.startswith("UD_"), "failed to obtain reference ID from mutalyzer"

    # cache the result
    chr_id_cache[(chrAcc, start, end)] = ref_id

    return ref_id


def apply_variants_mutalyzer(variants, chr_id):
    """
    Get the sequence obtained by applying the specified 'variants'
    to the reference sequence

    The variants to be applied must be specified with coordinates relative to
    the start of the region defined by chr_id

    :param variants: a list strings containing hgvs variant definitions
    :param chr_id: the mutalyzer id for the target reference region
    :return: A string containing the 'mutated' sequence
    """
    if len(variants) == 0:
        variants = ["1dup", "1del"] # this will give reference sequence

    mutalyzer_url = "{}/json/runMutalyzer".format(MUTALYZER)
    mutalyzer_params = {
        "variant": chr_id + ":g.[{}]".format(";".join(variants))
    }
    response = requests.get(url=mutalyzer_url, params=mutalyzer_params)
    sequence = response.json()

    return sequence["mutated"]


def apply_variants(variants, chrAcc, start, end):
    """
    Generate a 'mutated' sequence by applying the given variants to the specified genomic region

    :param variants: a list of strings containing the hgvs definitions of the variants to be applied
    :param chrAcc: Accession number for the target chromosome sequence
    :param start: start position of genomic region
    :param end: end position of genomic region
    :return: A string containing the 'mutated' sequence
    """
    chr_id = get_mutalyzer_region_id(chrAcc, start, end)
    return apply_variants_mutalyzer([offset_hgvs(v, -start + 1) for v in variants], chr_id)


def chrom_to_accession(chrom, build='hg19'):
    """
    Generate accession number from chromosome name
    :param chrom: name of chromosome
    :param build: The genome build
    :return: A string containing the accession number
    """
    mutalyzer_url = "{0}/json/chromAccession".format(MUTALYZER)
    mutalyzer_params = {
        'build': build,
        'name': chrom
    }
    response = requests.get(url=mutalyzer_url, params=mutalyzer_params)
    return response.json()



from setuptools import setup

setup(
    name="pgx-variant-tools",
    version="0.0.3",
    description="Tools for working with variants.",
    author="Guy Allard",
    author_email="wgallard@lumc.nl",
    url="https://github.com/LUMC/pgx-variant-tools",
    platforms=['any'],
    packages=["variant_tools"],
    install_requires=[
        'numpy==1.13.1',
        'biopython==1.69',
        'edlib==1.1.2.post2',
        'pyinterval==1.2.0'
    ],
    tests_requires=['pytest'],
    entry_points={
      "console_scripts": [
          "vcf2sequence = variant_tools.cli:vcf2sequence"
      ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='bioinformatics'
)



"""
Calling of variants using a minimal set of HGVS-like descriptions

The minimal notation uses three operators:
    substitution
    insertion
    deletion
"""
from enum import Enum


class Opcodes(Enum):
    """
    Operator codes
    """
    EQ  = 0
    SUB = 1
    INS = 2
    DEL = 3
    SKIP = 4


class Operator(object):
    """
    base class for all operators
    """
    opcode = None
    
    def __init__(self, pos=0, ref=None, alt=None):
        self.pos = pos
        
        self.ref = []
        if ref is not None:
            self.ref.append(ref)
            
        self.alt = []
        if alt is not None:
            self.alt.append(alt)

    def __str__(self):
        raise NotImplementedError()


class Substitution(Operator):
    """
    substitution operator
    """
    opcode = Opcodes.SUB
    
    def __str__(self):
        return "{}{}>{}".format(self.pos, self.ref[0], self.alt[0])
    
    
class Deletion(Operator):
    """
    deletion operator
    """
    opcode = Opcodes.DEL
    
    def __str__(self):
        if len(self.ref) == 1:
            return "{}del{}".format(self.pos, self.ref[0])
        
        return "{}_{}del{}".format(self.pos, self.pos + len(self.ref) - 1, "".join(self.ref))


class Insertion(Operator):
    """
    insertion operator
    """
    opcode = Opcodes.INS
    
    def __str__(self):
        return "{}_{}ins{}".format(self.pos, self.pos + 1, "".join(self.alt))

  
def get_opcode(ref_base, alt_base):
    """
    Get the opcode for the operation needed to convert a the given reference base into the given alt base
    
    :param ref_base: Character representing the reference base
    :param alt_base: Character representing the alternative base
    :return: 
    """
    if ref_base == alt_base:
        return Opcodes.EQ
    if ref_base == "$" or alt_base == "$":
        return Opcodes.SKIP
    if ref_base == "-":
        return Opcodes.INS
    if alt_base == "-":
        return Opcodes.DEL
    return Opcodes.SUB


def call_variants(alignment, offset=0):
    """
    Describe an alignment as a list of simple operations,
    where an operation is one of Substitution, Insertion or Deletion
    
    :param alignment: An alignment object tuple of the structure (ref_seq_aligned, alt_seq_aligned)
    :param offset:    An integer offset to be applied to all positions
    :return:          A list of variant operators. 
                      For compatability with HGVS, operators use a 1-based coordinate system
    """
    ops = []                # list of variant operations
    cur_op = None           # the current variant operation
    s_ref = alignment[0]    # ref sequence
    s_alt = alignment[1]    # alt sequence
    r_pos = -1              # current position on reference sequence
    
    # iterate over the alignment
    for i in range(len(alignment[0])):
        c_ref = s_ref[i]         # ref character at position i
        c_alt = s_alt[i]         # alt character at position i
        
        # increment ref sequence position
        if c_ref != "-":
            r_pos += 1
        
        # determine the opcode for ref vs alt at current position
        op = get_opcode(c_ref, c_alt)
        
        # store current op if new op is different
        if cur_op is not None and op != cur_op.opcode:
            ops.append(cur_op)
            cur_op = None
            
        # substitution: single-nucleotide operation
        # create Substitution operator, add it to ops
        if op == Opcodes.SUB:
            ops.append(Substitution(r_pos + 1 + offset, c_ref, c_alt))
        
        # insertion: multi-nucleotide operation,
        # create Insertion operator and hold onto it for further work
        if op == Opcodes.INS:
            if cur_op is None:
                cur_op = Insertion(r_pos + 1 + offset, alt=c_alt)
            else:
                cur_op.alt.append(c_alt)
        
        # deletion: multi-nucleotide operation,
        # create Deletion operator and hold onto it for further work
        if op == Opcodes.DEL:
            if cur_op is None:
                cur_op = Deletion(r_pos + 1 + offset, ref=c_ref)
            else:
                cur_op.ref.append(c_ref)
    
    # append the last operation
    if cur_op is not None:
        ops.append(cur_op)
    
    return ops
